///
/// \file TeaTimer.ino
/// \brief A Script to control a servo for building a tea timer with an Arduino board.
/// This sketch is implemented for use with the Arduino environment.
/// 
/// \author Matthias Hertel, http://www.mathertel.de
/// \copyright Copyright (c) 2014-2015 by Matthias Hertel.\n
/// This work is licensed under a BSD style license.\n
/// For more information, see: http://www.mathertel.de/License.aspx
/// 
/// \details
/// This sketch is part of my work on the Arduino platform at http://www.mathertel.de/Arduino
///
/// Hardware setup:
/// Connect the button to pin A0 and GND.
/// Connect a LED including a 470 Ohm resistor to pin D8 and GND.
/// Connect the servo to VCC (+5V), GND and pin D9.
/// You can change the used ports. see below.
///
/// History:
/// --------
/// * 01.03.2015 Created by Matthias Hertel
/// * 09.03.2015 Simple Arduino version working.
/// * 20.03.2015 Advanced Arduino version working.
/// --------

#include <Servo.h>

// TeaTimer state value definitions.
#define STATE_INIT  0x00  ///< Waiting for first button down
#define STATE_DOWN  0x10  ///< The button is pressed
#define STATE_UP    0x11  ///< The button is released so increment the tea-time by one.
#define STATE_ENTER 0x21  ///< entering the brewing phase by showing the number of minutes
#define STATE_BREW  0x22  ///< enter the tea bag and wait until end of brewing time
#define STATE_END   0x99  ///< brewing time ended.

// Some timing constants:
#define MSECS_TO_START   2000 ///< When button is not pressed since 2 seconds, the start brewing.
#define MSECS_FOR_DROPS 10000 ///< the number of seconds the bag is nearby over the tea pot.
#define MSECS_TO_DETACH   400  ///< after 400 msecs of no movement the servo is detached to save energy.

#define SECS_PER_MINUTE 60  ///< 60 minutes per second. for faster testing, reduce this to for example to 20.

// then dip the tea bag every minute
// from second 42 to 46
#define SECS_DIPSTART   42
#define SECS_DIPEND     46

/// The tea bag positions for the servo.
/// These values may depend on the mechanical layout of the TeaTimer and the used servo.
#define TEABAG_INITIAL 95
#define TEABAG_ENTER 65
#define TEABAG_DIP 70
#define TEABAG_OUT 80

// IO ports definitions
#define BUTTONPIN A0 /// The button is attached to pin 0 
#define LEDPIN     8 // The LED is attached to pin # 8
#define SERVOPIN   9 // Servo is attached to pin # 9


Servo myServo;  //create a servo object.

// global variables

uint8_t state; ///< state of the tea timer script.
uint8_t teaTime; ///< duration of brewing in seconds.

unsigned long started; ///< the time when the current state started.

// The servo is not ordered to move in one step to make a slow movement.
bool tbAttached;
int tbPosition;      ///< the current position of the tea bag holder.
int tbPositionNext;  ///< the next position.
unsigned long tbLastMove;


/// This function is used to move the servo to a new position.
/// There is no need to keep the servo attached all the time because the mechanical design
/// of the servo allows to hold a tea bag in place without actively controlling the servo.
/// I found that 200 msecs waiting at the end of a movement is enough
/// and the detaching the servo so it doesn't drain the battery too much.
/// \return Return true, when target position is reached.
bool tbCheckPosition() {
  unsigned long now = millis();
  bool ret = true; // assume the position is ok.

  if (tbPosition == tbPositionNext) {
    // Detach the servo, when the final position is reached and some time has passed.
    if ((tbAttached == true) && (now > tbLastMove + MSECS_TO_DETACH)) {
      myServo.detach();
      tbAttached = false;
    } // if

  } else if (tbPosition != tbPositionNext) {
    ret = false; // position is not ok yet.
    // Attach the servo, when the position has to be changed and not attached yet.
    if (tbAttached == false) {
      myServo.attach(SERVOPIN);        // Attach the servo, make it active
      tbAttached = true;
    } // if

    // move the servo to the new position s l o w l y
    if (tbPosition < tbPositionNext) {
      tbPosition += 1;
    } else {
      tbPosition -= 1;
    } // if

    myServo.write(tbPosition);
    delay(50);
    tbLastMove = now;
  } // if

  return (ret);
} // tbCheckPosition()


/// Set a new target position of the tea timer servo.
/// \param nextPos
/// \param wait
void tbSetPosition(int position, bool wait = false) {
  tbPositionNext = position;
  if (wait) {
    while (! tbCheckPosition()) {
      // wait ...
    };
  } // if

} // tbSetPosition()


/// Initialize all hardware & software.
void setup()
{
  // Setup IO
  pinMode(LEDPIN, OUTPUT);  // Setting LED pin as output
  pinMode(BUTTONPIN, INPUT);  // ???
  digitalWrite(BUTTONPIN, HIGH); // internal pull Up

  // flash the LED to show that the battery is having power
  digitalWrite(LEDPIN, HIGH);

  // make servo move to starting position fast.
  tbPositionNext = tbPosition = TEABAG_INITIAL;
  tbAttached = false;
  myServo.attach(SERVOPIN);        // Attach the servo, make it active
  myServo.write(TEABAG_INITIAL);
  delay(300);
  myServo.detach();

  digitalWrite(LEDPIN, LOW);

  // start with the initialize state.
  state = STATE_INIT;
} // setup()


/// check for the button, the time and the state
/// and do the servo and led...
void loop()
{
  unsigned long now = millis();
  unsigned long duration = now - started;
  uint8_t nextState = state; // no new state until here.

  tbCheckPosition();

  // get the current signal from the button.
  int b = digitalRead(BUTTONPIN);

  if (state == STATE_INIT) {
    // When nothing happened yet...

    if (b == LOW) {
      // The button signal got low, so the button was pressed for the first time.
      nextState = STATE_DOWN;
      teaTime = 0; // reset tea brewing time.
      digitalWrite(LEDPIN, HIGH); // switch the LED on
    }

  } else if (state == STATE_DOWN) {
    // When the button is down...
    if (b == HIGH) {
      // the button signal got high, so the button was released.
      nextState = STATE_UP;
      digitalWrite(LEDPIN, LOW); // switch the LED off

      // only increment the teaTime when the key press was real and not a bouncing)
      if (duration > 50) {
        teaTime += 1;
      } // if
    } // if

  } else if (state == STATE_UP) {
    // When the button is up...

    if (b == LOW) {
      // The button signal got low, so the button was pressed another time.
      nextState = STATE_DOWN;
      digitalWrite(LEDPIN, HIGH);

    } else if (duration > MSECS_TO_START) {
      // some time (MSECS_TO_START) has passed since the last button was pressed
      // so brewing can start.
      nextState = STATE_ENTER;

    } // if

  } else if (state == STATE_ENTER) {
    // When starting the brewing we have to entering the tea bag.

    // show the timer count:
    for (int n = 0; n < teaTime; n++) {
      digitalWrite(LEDPIN, HIGH);
      delay(200);
      digitalWrite(LEDPIN, LOW);
      delay(200);
    } // for

    // move to new position and wait for moving is over
    tbSetPosition(TEABAG_ENTER, true);
    nextState = STATE_BREW;

  } else if (state == STATE_BREW) {
    // When brewing...
    uint16_t durationSecs = duration / 1000; // duration in  seconds to avoid div by 1000 multiple times.

    // .. the LED should show activity.
    if ((duration % 1000) < 200) {
      // it's on during the first part of a second.
      digitalWrite(LEDPIN, HIGH);
    } else {
      // it's of the rest of the second.
      digitalWrite(LEDPIN, LOW);
    } // if

    // dip the bag a little bit around from time to time
    // moving to the dip position during second 51 and 60 of each minute
    if ((durationSecs % SECS_PER_MINUTE) == SECS_DIPSTART)
      tbSetPosition(TEABAG_DIP);
    if ((durationSecs % SECS_PER_MINUTE) == SECS_DIPEND)
      tbSetPosition(TEABAG_ENTER);

    if (durationSecs >= (teaTime * SECS_PER_MINUTE)) {
      // The waiting is over !
      digitalWrite(LEDPIN, HIGH);
      tbSetPosition(TEABAG_OUT);
      nextState = STATE_END;
    } // if

  } else if (state == STATE_END) {
    if (duration >= MSECS_FOR_DROPS) {
      // Start again...
      nextState = STATE_INIT;
      tbSetPosition(TEABAG_INITIAL);
      digitalWrite(LEDPIN, LOW);
    } // if

  } // if (state)

  if (nextState != state) {
    state = nextState;
    started = millis();
  } // if

} // loop()


// End.

